#include <HsFFI.h>
#if defined(__cplusplus)
extern "C" {
#endif
extern HsInt32 gtk2hs_store_get_n_columns_impl(HsStablePtr a1);
extern HsWord64 gtk2hs_store_get_column_type_impl(HsStablePtr a1, HsInt32 a2);
extern HsInt32 gtk2hs_store_get_flags_impl(HsStablePtr a1);
extern HsInt32 gtk2hs_store_get_iter_impl(HsStablePtr a1, HsPtr a2, HsPtr a3);
extern HsPtr gtk2hs_store_get_path_impl(HsStablePtr a1, HsPtr a2);
extern void gtk2hs_store_get_value_impl(HsStablePtr a1, HsPtr a2, HsInt32 a3, HsPtr a4);
extern HsInt32 gtk2hs_store_iter_next_impl(HsStablePtr a1, HsPtr a2);
extern HsInt32 gtk2hs_store_iter_children_impl(HsStablePtr a1, HsPtr a2, HsPtr a3);
extern HsInt32 gtk2hs_store_iter_has_child_impl(HsStablePtr a1, HsPtr a2);
extern HsInt32 gtk2hs_store_iter_n_children_impl(HsStablePtr a1, HsPtr a2);
extern HsInt32 gtk2hs_store_iter_nth_child_impl(HsStablePtr a1, HsPtr a2, HsPtr a3, HsInt32 a4);
extern HsInt32 gtk2hs_store_iter_parent_impl(HsStablePtr a1, HsPtr a2, HsPtr a3);
extern void gtk2hs_store_ref_node_impl(HsStablePtr a1, HsPtr a2);
extern void gtk2hs_store_unref_node_impl(HsStablePtr a1, HsPtr a2);
extern HsInt32 gtk2hs_store_row_draggable_impl(HsPtr a1, HsStablePtr a2, HsPtr a3);
extern HsInt32 gtk2hs_store_drag_data_get_impl(HsPtr a1, HsStablePtr a2, HsPtr a3, HsPtr a4);
extern HsInt32 gtk2hs_store_drag_data_delete_impl(HsPtr a1, HsStablePtr a2, HsPtr a3);
extern HsInt32 gtk2hs_store_drag_data_received_impl(HsPtr a1, HsStablePtr a2, HsPtr a3, HsPtr a4);
extern HsInt32 gtk2hs_store_row_drop_possible_impl(HsPtr a1, HsStablePtr a2, HsPtr a3, HsPtr a4);
#if defined(__cplusplus)
}
#endif

